# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column classes that display information in peer lists"""

import urwid

from ..table import ColumnHeaderWidget
from . import (Style, CellWidgetBase)
from ...views.plist import COLUMNS as _COLUMNS


TUICOLUMNS = {}

class TorrentName(_COLUMNS['name'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='peerlist.name', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['name'].header),
                           style.attrs('header'))

TUICOLUMNS['name'] = TorrentName


class Client(_COLUMNS['client'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='peerlist.client', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['client'].header),
                           style.attrs('header'))

TUICOLUMNS['client'] = Client


class Country(_COLUMNS['country'], CellWidgetBase):
    width = 7
    style = Style(prefix='peerlist.country', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['country'].header),
                           style.attrs('header'))

TUICOLUMNS['country'] = Country


class IPAddress(_COLUMNS['ip'], CellWidgetBase):
    style = Style(prefix='peerlist.ip', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['ip'].header),
                           style.attrs('header'))

TUICOLUMNS['ip'] = IPAddress


class Port(_COLUMNS['port'], CellWidgetBase):
    style = Style(prefix='peerlist.port', focusable=False,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['port'].header),
                           style.attrs('header'))

TUICOLUMNS['port'] = Port


class Progress(_COLUMNS['progress'], CellWidgetBase):
    style = Style(prefix='peerlist.progress', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['progress'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else None

TUICOLUMNS['progress'] = Progress


class RateDown(_COLUMNS['rate-down'], CellWidgetBase):
    style = Style(prefix='peerlist.rate-down', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-down'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else None

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-down'].header['right']

TUICOLUMNS['rate-down'] = RateDown


class RateUp(_COLUMNS['rate-up'], CellWidgetBase):
    style = Style(prefix='peerlist.rate-up', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-up'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else None

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-up'].header['right']

TUICOLUMNS['rate-up'] = RateUp


class ETA(_COLUMNS['eta'], CellWidgetBase):
    style = Style(prefix='peerlist.eta', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['eta'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if bool(self.value) else None

TUICOLUMNS['eta'] = ETA


class EstimatedPeerRate(_COLUMNS['rate-est'], CellWidgetBase):
    style = Style(prefix='peerlist.rate-est', focusable=False,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['rate-est'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value > 0 else None

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['rate-est'].header['right']

TUICOLUMNS['rate-est'] = EstimatedPeerRate



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Column classes that display information in torrent file lists"""

import urwid

from ..table import ColumnHeaderWidget
from . import (Style, CellWidgetBase)
from ...views.flist import COLUMNS as _COLUMNS


TUICOLUMNS = {}


from .common_columns import MarkedBase
class Marked(MarkedBase):
    style = Style(prefix='filelist.marked', focusable=True, extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(), style.attrs('header'))

TUICOLUMNS['marked'] = Marked


class Filename(_COLUMNS['name'], CellWidgetBase):
    width = ('weight', 100)
    style = Style(prefix='filelist.name', focusable=True,
                  extras=('header',), modes=('file', 'folder', 'filtered'))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['name'].header),
                           style.attrs('header'))

    def get_mode(self):
        if isinstance(self.data['id'], int):
            return 'file'
        else:
            return 'folder'

TUICOLUMNS['name'] = Filename


class Size(_COLUMNS['size'], CellWidgetBase):
    style = Style(prefix='filelist.size', focusable=True,
                  extras=('header',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['size'].header),
                           style.attrs('header'))

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['size'].header['right']

TUICOLUMNS['size'] = Size


class Downloaded(_COLUMNS['downloaded'], CellWidgetBase):
    style = Style(prefix='filelist.downloaded', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['downloaded'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.data['progress'] < 100 else None

    @classmethod
    def set_unit(cls, unit):
        cls.header.original_widget.right = _COLUMNS['downloaded'].header['right']

TUICOLUMNS['downloaded'] = Downloaded


class Progress(_COLUMNS['progress'], CellWidgetBase):
    style = Style(prefix='filelist.progress', focusable=True,
                  extras=('header',), modes=('highlighted',))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['progress'].header),
                           style.attrs('header'))

    def get_mode(self):
        return 'highlighted' if self.value < 100 else ''

TUICOLUMNS['progress'] = Progress


class Priority(_COLUMNS['priority'], CellWidgetBase):
    style = Style(prefix='filelist.priority', focusable=True,
                  extras=('header',), modes=('low', 'high', 'shun'))
    header = urwid.AttrMap(ColumnHeaderWidget(**_COLUMNS['priority'].header),
                           style.attrs('header'))

    def get_mode(self):
        val = self.value
        return val if val in ('low', 'high', 'shun') else None

TUICOLUMNS['priority'] = Priority



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


class ConstantBase():
    pass

def is_constant(obj):
    return isinstance(obj, ConstantBase)

_constants_cache = {}
def get_constant(string, repr_str=None, bases=(), value=None, attrs={}):
    """Get constant/singleton defined by `string`

    repr_str: Same as `string.upper()` if not specified
    bases: tuple of base classes; will be prepended to (ConstantBase, str)
    value: What the constant class is passed to upon creation
    attrs: Additional attributes of the constant class
    """
    if string not in _constants_cache:
        repr_str = (string if repr_str is None else repr_str).upper()

        def __str__(self): return self.string
        def __repr__(self): return '<Constant: {}>'.format(self.repr_str.upper())

        cls_attrs = {'__str__': __str__,
                     '__repr__': __repr__,
                     'string': string,
                     'repr_str': repr_str,
                     'value': value}
        cls_attrs.update(attrs)

        cls = type('Constant', bases + (ConstantBase,), cls_attrs)

        if value is not None:
            _constants_cache[string] = cls(value)
        else:
            _constants_cache[string] = cls()
    return _constants_cache[string]


DISCONNECTED = get_constant('disconnected')
UNLIMITED = get_constant('unlimited', bases=(float,), value='inf')
DISABLED = get_constant('disabled', attrs={'__bool__': lambda self: False})
ENABLED = get_constant('enabled', attrs={'__bool__': lambda self: True})



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid

from ..main import srvapi
from ..main import cfg
from . import main as tui
from ..client import constants as const


class KeyChainsWidget(urwid.WidgetWrap):
    def __init__(self):
        self._pile = urwid.Pile([])
        super().__init__(urwid.AttrMap(self._pile, 'keychains'))

    def update(self, keychains):
        self._pile.contents[:] = []
        if not keychains:
            return

        lines = []
        keys_col_width = 0
        for kc,action in sorted(keychains, key=lambda x: str(x[0]).lower()):
            given = ' '.join(kc.given)
            current = kc.next_key
            rest = ' '.join(kc[len(kc.given)+1:])
            keys_col_width = max(len(' '.join((given, current, rest))), keys_col_width)
            lines.append((given, current, rest, str(action)))

        for given,current,rest,action in lines:
            keys_widget = urwid.Text([('keychains.keys', given), ' ',
                                      ('keychains.keys.next', current), ' ',
                                      ('keychains.keys', rest)])
            action_widget = urwid.Text(('keychains.action', action))
            line_widget = urwid.Columns([(keys_col_width, keys_widget),
                                         ('pack', urwid.Text('  -  ')),
                                         action_widget])
            self._pile.contents.append((line_widget, self._pile.options('pack')))

    def rows(self, size, focus=False):
        return len(self._pile.contents)

    _selectable = False
    _sizing = ['flow']


class QuickHelpWidget(urwid.Text):
    def __init__(self):
        super().__init__('')

    def update(self):
        def get_key(cmd, contexts):
            """Return shortest key sequence that executes `cmd`"""
            for context in contexts:
                keys = tuple(tui.keymap.keys(lambda key,action: action.startswith(cmd), context))
                if keys:
                    return sorted(keys, key=lambda k: len(k))[0]

        def get_first_key(cmd, contexts):
            """Same as `get_key`, but return only first key if key sequence"""
            key = get_key(cmd, contexts)
            if isinstance(key, tuple):  # key is a key sequence
                return key[0]
            else:
                return key

        def maybe_add_entry(items, label, key):
            if key is not None:
                items.append([('topbar.help.space',  '   '),
                              ('topbar.help.key',    str(key)),
                              ('topbar.help.equals', ' '),
                              ('topbar.help.label',  label)])

        items = []
        maybe_add_entry(items, 'Prompt', get_key('tui show cli', contexts=('main', None)))
        maybe_add_entry(items, 'Quit', get_key('quit', contexts=('main', None)))
        maybe_add_entry(items, 'Help', get_first_key('tab help', contexts=('main', None)))
        self.set_text(items) if items else self.set_text('')


class ConnectionStatusWidget(urwid.WidgetWrap):
    def __init__(self):
        self._text = urwid.Text('Not connected')
        self._attrmap = urwid.AttrMap(self._text, 'topbar.host.disconnected')
        super().__init__(self._attrmap)
        srvapi.rpc.on('connected', self._handle_connected)
        srvapi.rpc.on('disconnected', self._handle_disconnected)
        srvapi.rpc.on('error', self._handle_error)

    def _handle_connected(self, url):
        self._text.set_text('{}:{} Transmission {}'.format(url.hostname, url.port, srvapi.rpc.version))
        self._attrmap.set_attr_map({None: 'topbar.host.connected'})

    def _handle_disconnected(self, url):
        self._text.set_text(str(url))
        self._attrmap.set_attr_map({None: 'topbar.host.disconnected'})

    def _handle_error(self, url, error):
        self._text.set_text(str(error))  # error should also contain url
        self._attrmap.set_attr_map({None: 'topbar.host.disconnected'})


class BandwidthStatusWidget(urwid.Widget):
    _RATE_WIDTH = 6
    _SPACER_WIDTH = 1
    _TAIL_WIDTH = 4

    def __init__(self):
        self._text_up = urwid.Text('', align='right')
        self._text_up_limit = urwid.Text('')
        self._text_dn = urwid.Text('', align='right')
        self._text_dn_limit = urwid.Text('')

        spacer = urwid.AttrMap(urwid.SolidFill(' '), 'bottombar')
        self._spacer_canvas = spacer.render((self._SPACER_WIDTH, 1))

        def mkattr(text, attrsname):
            return urwid.AttrMap(urwid.Padding(text), attrsname)

        self._attr_up = mkattr(self._text_up, 'bottombar.bandwidth.up')
        self._attr_up_limit = mkattr(self._text_up_limit, 'bottombar.bandwidth.up')
        self._attr_dn = mkattr(self._text_dn, 'bottombar.bandwidth.down')
        self._attr_dn_limit = mkattr(self._text_dn_limit, 'bottombar.bandwidth.down')

        self._up_limit_width = 0
        self._dn_limit_width = 0
        self._connected = False

        srvapi.status.on_update(self._update_current_rates)
        srvapi.settings.on_update(self._update_rate_limits)

    def _update_current_rates(self, status):
        highlight = lambda v: v > 0

        up = status.rate_up
        if up is not const.DISCONNECTED:
            up_attrs = 'bottombar.bandwidth.up.highlighted' if highlight(up) \
                       else 'bottombar.bandwidth.up'
            self._text_up.set_text((up_attrs, str(up)))

        dn = status.rate_down
        if dn is not const.DISCONNECTED:
            dn_attrs = 'bottombar.bandwidth.down.highlighted' if highlight(dn) \
                       else 'bottombar.bandwidth.down'
            self._text_dn.set_text((dn_attrs, str(dn)))
            self._connected = True
        else:
            self._connected = False
        self._invalidate()

    def _update_rate_limits(self, settings):
        up = settings['rate-limit-up']
        dn = settings['rate-limit-down']

        # Empty display if DISCONNECTED or UNLIMITED
        as_str = lambda v: '' if v in (const.DISCONNECTED, const.UNLIMITED) \
                           else '/%s' % str(v)
        self._text_up_limit.set_text(as_str(up))
        self._text_dn_limit.set_text(as_str(dn))

        self._up_limit_width = len(self._text_up_limit.text)
        self._dn_limit_width = len(self._text_dn_limit.text)
        self._invalidate()

    def _mk_tail_canv(self, direction, icon):
        unit = {'bit': 'b', 'byte': 'B'}[cfg['unit.bandwidth'].value]
        text = urwid.Text('%s/s%s' % (unit, icon))
        attr_text = urwid.AttrMap(text, 'bottombar.bandwidth.%s' % direction)
        return attr_text.render((self._TAIL_WIDTH,))

    def render(self, size, focus=False):
        if not self._connected:
            size = self.pack(size)
            return urwid.AttrMap(urwid.SolidFill(' '), 'bottombar').render(size)

        canv_up_tail = self._mk_tail_canv('up', '↑')
        canv_dn_tail = self._mk_tail_canv('down', '↓')

        # List of (canvas, position, focus) tuples
        combinelist = []

        # Download
        canv_dn = self._attr_dn.render((self._RATE_WIDTH,))
        combinelist.append((canv_dn, None, False, self._RATE_WIDTH))
        if self._dn_limit_width > 0:
            canv_dn_limit = self._attr_dn_limit.render((self._dn_limit_width,))
            combinelist.append((canv_dn_limit, None, False, self._dn_limit_width))
        combinelist.append((canv_dn_tail, None, False, self._TAIL_WIDTH))

        combinelist.append((self._spacer_canvas, None, False, self._SPACER_WIDTH))

        # Upload
        canv_up = self._attr_up.render((self._RATE_WIDTH,))
        combinelist.append((canv_up, None, False, self._RATE_WIDTH))
        if self._up_limit_width > 0:
            canv_up_limit = self._attr_up_limit.render((self._up_limit_width,))
            combinelist.append((canv_up_limit, None, False, self._up_limit_width))
        combinelist.append((canv_up_tail, None, False, self._TAIL_WIDTH))

        return urwid.CanvasJoin(combinelist)

    _sizing = ['fixed']

    def rows(self, size, focus=False):
        return 1

    def pack(self, size, focus=False):
        cols = (self._RATE_WIDTH*2) + (self._TAIL_WIDTH*2) + \
               self._dn_limit_width + self._up_limit_width + \
               self._SPACER_WIDTH
        return (cols, 1)


class TorrentCountersWidget(urwid.WidgetWrap):
    def __init__(self):
        self._text = urwid.Text(' ')
        super().__init__(urwid.AttrMap(self._text, 'bottombar'))
        tui.srvapi.status.on_update(self._update_counters)

    def _update_counters(self, status):
        counters = status.count
        if counters.total is const.DISCONNECTED:
            self._text.set_text('')
            return

        parts = []
        for name in ('downloading', 'uploading', 'stopped'):
            count = getattr(counters, name)
            if count > 0:
                parts.append('%d %s' % (count, name))

        text = ['{} torrents'.format(str(counters.total))]
        if parts:
            text[-1] += ': ' + ', '.join(parts)

        if counters.isolated > 0:
            text[-1] += ', '
            text.append(('bottombar.important', '%s isolated' % counters.isolated))
        self._text.set_text(text)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering Torrents by their values"""

from ...logging import make_logger
log = make_logger(__name__)

import re

from . import (BoolFilterSpec, CmpFilterSpec, make_cmp_filter,
               Filter, FilterChain)

from ..tkeys import TYPES as VALUETYPES
def _make_cmp_filter(*args, **kwargs):
    return make_cmp_filter(VALUETYPES, *args, **kwargs)

def _desc(text):
    if text.startswith('...'):
        text = 'Match VALUE against ' + text[4:]
    return text


from ..tkeys import Status
_STATUS_VERIFY    = Status.VERIFY
_STATUS_DOWNLOAD  = Status.DOWNLOAD
_STATUS_UPLOAD    = Status.UPLOAD
_STATUS_INIT      = Status.INIT
_STATUS_CONNECTED = Status.CONNECTED
_STATUS_ISOLATED  = Status.ISOLATED
_STATUS_QUEUED    = Status.QUEUED
_STATUS_SEED      = Status.SEED
_STATUS_IDLE      = Status.IDLE
_STATUS_STOPPED   = Status.STOPPED

class SingleTorrentFilter(Filter):
    DEFAULT_FILTER = 'name'

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda t: True,
            description='All torrents',
            needed_keys=(),
            aliases=('*',)),

        'leeching': BoolFilterSpec(
            lambda t: t['%downloaded'] < 100 and _STATUS_STOPPED not in t['status'],
            description='unstopped Torrents downloading or waiting for seeds',
            needed_keys=('%downloaded', 'status')),
        'seeding': BoolFilterSpec(
            lambda t: t['%downloaded'] >= 100 and _STATUS_STOPPED not in t['status'],
            description='unstopped Torrents with all wanted files downloaded',
            needed_keys=('%downloaded', 'status')),
        'complete': BoolFilterSpec(
            lambda t: t['%downloaded'] >= 100,
            description='Torrents with all wanted files downloaded',
            needed_keys=('%downloaded',)),
        'incomplete': BoolFilterSpec(
            lambda t: t['%downloaded'] < 100,
            description='Torrents with some wanted files not fully downloaded',
            needed_keys=('%downloaded',)),
        'stopped': BoolFilterSpec(
            lambda t: _STATUS_STOPPED in t['status'],
            description='Torrents not allowed to up- or download',
            needed_keys=('status',),
            aliases=('paused',)),

        'active': BoolFilterSpec(
            lambda t: t['peers-connected'] > 0 or _STATUS_VERIFY in t['status'],
            description='Torrents connected to peers or being verified',
            needed_keys=('peers-connected', 'status')),
        'downloading': BoolFilterSpec(
            lambda t: t['rate-down'] > 0,
            description='Torrents using download bandwidth',
            needed_keys=('rate-down',)),
        'uploading': BoolFilterSpec(
            lambda t: t['rate-up'] > 0,
            description='Torrents using upload bandwidth',
            needed_keys=('rate-up',)),
        'verifying': BoolFilterSpec(
            lambda t: _STATUS_VERIFY in t['status'],
            description='Torrents being verified or queued for verification',
            needed_keys=('status',),
            aliases=('checking',)),
        'idle': BoolFilterSpec(
            lambda t: (_STATUS_IDLE in t['status'] and
                       _STATUS_STOPPED not in t['status']),
            description='unstopped Torrents not using any bandwidth',
            needed_keys=('status',)),
        'isolated': BoolFilterSpec(
            lambda t: _STATUS_ISOLATED in t['status'],
            description='Torrents that cannot discover new peers in any way',
            needed_keys=('status',)),

        'private': BoolFilterSpec(
            lambda t: t['private'],
            description='Torrents connectable through trackers only',
            needed_keys=('private',)),
        'public': BoolFilterSpec(
            lambda t: not t['private'],
            description='Torrents connectable through DHT and/or PEX',
            needed_keys=('private',)),
    }


    # Filters with arguments
    COMPARATIVE_FILTERS = {
        'connections': _make_cmp_filter('peers-connected', _desc('... number of connected peers'),
                                        aliases=('conn',)),
        '%downloaded': _make_cmp_filter('%downloaded', _desc('... percentage of downloaded bytes'),
                                        aliases=('%down',)),
        'downloaded': _make_cmp_filter('size-downloaded', _desc('... number of downloaded bytes'),
                                       aliases=('down',)),
        'id':        _make_cmp_filter('id', _desc('... ID')),
        'name':      _make_cmp_filter('name', _desc('... name'),
                                      aliases=('title',)),
        'path':      _make_cmp_filter('path', _desc('... full path to download directory'),
                                      aliases=('dir',)),
        'ratio':     _make_cmp_filter('ratio', _desc('... uploaded/downloaded ratio')),
        'rate-down': _make_cmp_filter('rate-down', _desc('... download rate'),
                                      aliases=('rdown',)),
        'rate-up':   _make_cmp_filter('rate-up', _desc('... upload rate'),
                                      aliases=('rup',)),
        'seeds':     _make_cmp_filter('peers-seeding',
                                      _desc('... largest number of seeds reported by any tracker')),
        'size':      _make_cmp_filter('size-final', _desc('... combined size of all wanted files')),
        'uploaded':  _make_cmp_filter('size-uploaded',
                                      _desc('... number of uploaded bytes')),

        'tracker': CmpFilterSpec(
            lambda t, op, v: any(op(tracker['url-announce'].domain, v)
                                 for tracker in t['trackers']),
            description=_desc('... domain of the announce URL of trackers'),
            needed_keys=('trackers',),
            value_type=str,
        ),

        'eta': CmpFilterSpec(
            lambda t, op, v: t['timespan-eta'].is_known and op(t['timespan-eta'], v),
            description=_desc('... estimated time for torrent to finish'),
            needed_keys=('timespan-eta',),
            value_type=VALUETYPES['timespan-eta'],
            value_convert=VALUETYPES['timespan-eta'].from_string,
        ),
        'created': CmpFilterSpec(
            lambda t, op, v: t['time-created'].is_known and op(t['time-created'], v),
            description=_desc('... time torrent was created'),
            needed_keys=('time-created',),
            value_type=VALUETYPES['time-created'],
            value_convert=VALUETYPES['time-created'].from_string,
        ),
        'added': CmpFilterSpec(
            lambda t, op, v: t['time-added'].is_known and op(t['time-added'], v),
            description=_desc('... time torrent was added'),
            needed_keys=('time-added',),
            value_type=VALUETYPES['time-added'],
            value_convert=VALUETYPES['time-added'].from_string,
        ),
        'started': CmpFilterSpec(
            lambda t, op, v: t['time-started'].is_known and op(t['time-started'], v),
            description=_desc('... last time torrent was started'),
            needed_keys=('time-started',),
            value_type=VALUETYPES['time-started'],
            value_convert=VALUETYPES['time-started'].from_string,
        ),
        'activity': CmpFilterSpec(
            lambda t, op, v: t['time-activity'].is_known and op(t['time-activity'], v),
            description=_desc('... last time torrent was active'),
            needed_keys=('time-activity',),
            value_type=VALUETYPES['time-activity'],
            value_convert=VALUETYPES['time-activity'].from_string,
        ),
        'completed': CmpFilterSpec(
            lambda t, op, v: t['time-completed'].is_known and op(t['time-completed'], v),
            description=_desc('... time all wanted files where downloaded'),
            needed_keys=('time-completed',),
            value_type=VALUETYPES['time-completed'],
            value_convert=VALUETYPES['time-completed'].from_string,
        ),
    }



class TorrentFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTorrentFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ...logging import make_logger
log = make_logger(__name__)

from . import (SortSpecBase, SorterBase)

class _SortSpec(SortSpecBase):
    def __init__(self, *args, description='', needed_keys=(), **kwargs):
        description = 'Sort torrents by %s' % description
        super().__init__(*args, description=description, **kwargs)
        self.needed_keys = needed_keys


class TorrentSorter(SorterBase):
    SORTSPECS = {
        'name':         _SortSpec(lambda t: t['name'].lower(),
                                  needed_keys=('name',),
                                  description='name'),
        'path':         _SortSpec(lambda t: t['path'],
                                  aliases=('dir',),
                                  needed_keys=('path',),
                                  description='download location'),
        'status':       _SortSpec(lambda t: t['status'],
                                  needed_keys=('status',),
                                  description='current status (idle, uploading, verifying, etc.)'),
        'size':         _SortSpec(lambda t: t['size-final'],
                                  needed_keys=('size-final',),
                                  description='number of bytes of all wanted files'),
        'peers':        _SortSpec(lambda t: t['peers-connected'],
                                  needed_keys=('peers-connected',),
                                  description='connected peers'),
        'seeds':        _SortSpec(lambda t: t['peers-seeding'],
                                  needed_keys=('peers-seeding',),
                                  description='highest number of seeds reported by any tracker'),
        'ratio':        _SortSpec(lambda t: t['ratio'],
                                  needed_keys=('ratio',),
                                  description='upload/download ratio'),
        'rate-down':    _SortSpec(lambda t: t['rate-down'],
                                  needed_keys=('rate-down',),
                                  description='download rate'),
        'rate-up':      _SortSpec(lambda t: t['rate-up'],
                                  needed_keys=('rate-up',),
                                  description='upload rate'),
        'rate':         _SortSpec(lambda t: t['rate-up'] + t['rate-down'],
                                  needed_keys=('rate-up', 'rate-down'),
                                  description='combined download and upload rate'),
        'uploaded':     _SortSpec(lambda t: t['size-uploaded'],
                                  needed_keys=('size-uploaded',),
                                  description='number of uploaded bytes'),
        'downloaded':   _SortSpec(lambda t: t['size-downloaded'],
                                  needed_keys=('size-downloaded',),
                                  description='number of downloaded bytes'),
        'progress':     _SortSpec(lambda t: t['%downloaded'],
                                  lambda t: t['%metadata'],
                                  lambda t: t['%verified'],
                                  needed_keys=('%downloaded', '%metadata', '%verified'),
                                  description='downloading or verifying progress'),
        'tracker':      _SortSpec(lambda t: t['trackers'][0]['url-announce'].domain if t['trackers'] else '',
                                  needed_keys=('trackers',),
                                  description='domain of first tracker'),
        'eta':          _SortSpec(lambda t: t['timespan-eta'],
                                  needed_keys=('timespan-eta',),
                                  description='estimated time to finish downloading'),
        'created':      _SortSpec(lambda t: t['time-created'],
                                  needed_keys=('time-created',),
                                  description='creation time'),
        'added':        _SortSpec(lambda t: t['time-added'],
                                  needed_keys=('time-added',),
                                  description='time of addition'),
        'started':      _SortSpec(lambda t: t['time-started'],
                                  needed_keys=('time-started',),
                                  description='start time'),
        'activity':     _SortSpec(lambda t: t['time-activity'],
                                  needed_keys=('time-activity',),
                                  description='time of latest upload/download activity'),
        'completed':    _SortSpec(lambda t: t['time-completed'],
                                  needed_keys=('time-completed',),
                                  description='time of completion'),
    }
    DEFAULT_SORT = 'name'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Torrent keys we need for all sortspecs
        self._needed_keys = tuple(set().union(
            *(sortspec.needed_keys
              for sortspec in self._sortspecs)
        ))

    @property
    def needed_keys(self):
        return self._needed_keys



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Filtering PeerList items by various values"""

from ..tkeys import TorrentPeer
from . import (BoolFilterSpec, CmpFilterSpec, Filter, FilterChain)


class SingleTorrentPeerFilter(Filter):
    DEFAULT_FILTER = None

    # Filters without arguments
    BOOLEAN_FILTERS = {
        'all': BoolFilterSpec(
            lambda p: True,
            aliases=('*',),
            description='All peers'),
        'uploading': BoolFilterSpec(
            lambda p: p['rate-up'] > 0,
            aliases=('ul',),
            description='Peers we are uploading to'),
        'downloading': BoolFilterSpec(
            lambda p: p['rate-down'] > 0,
            aliases=('dl',),
            description='Peers we are downloading from'),
        'seeding': BoolFilterSpec(
            lambda p: p['progress'] >= 100,
            aliases=('done',),
            description='Peers that have downloaded all data'),
    }

    COMPARATIVE_FILTERS = {
        'client': CmpFilterSpec(
            lambda p, op, v: op(p['client'], v),
            description='Match VALUE against peer client',
            value_type=TorrentPeer.TYPES['client']),
        'country': CmpFilterSpec(
            lambda p, op, v: op(p['country'], v),
            description='Match VALUE against peer country',
            value_type=TorrentPeer.TYPES['country']),
        'ip': CmpFilterSpec(
            lambda p, op, v: op(p['ip'], v),
            description='Match VALUE against peer IP address',
            value_type=TorrentPeer.TYPES['ip']),
        'port': CmpFilterSpec(
            lambda p, op, v: op(p['port'], v),
            description='Match VALUE against peer port',
            value_type=TorrentPeer.TYPES['port']),
        'downloaded': CmpFilterSpec(
            lambda p, op, v: op(p['tsize'] * (p['progress']/100), v),
            aliases=('down',),
            description='Match VALUE against number of bytes peer has downloaded',
            value_type=TorrentPeer.TYPES['tsize']),
        '%downloaded': CmpFilterSpec(
            lambda p, op, v: op(p['progress'], v),
            aliases=('%down',),
            description='Match VALUE against percentage of bytes peer has downloaded',
            value_type=TorrentPeer.TYPES['progress']),
    }


class TorrentPeerFilter(FilterChain):
    """One or more filters combined with & and | operators"""
    filterclass = SingleTorrentPeerFilter



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid

from .. import main as tui

from ..scroll import ScrollBar
from ..table import Table
from .plist_columns import TUICOLUMNS


class PeerListItemWidget(urwid.WidgetWrap):
    def __init__(self, peer, cells):
        self._peer = peer
        self._cells = cells
        self.update(peer)
        super().__init__(urwid.AttrMap(cells, 'peerlist'))

    def update(self, peer):
        for widget in self._cells.original_widget.widgets:
            widget.update(peer)
        self._peer = peer

    @property
    def pid(self):
        return self._peer['id']

    @property
    def peer(self):
        return self._peer


class PeerListWidget(urwid.WidgetWrap):
    def __init__(self, srvapi, tfilter, pfilter, columns, sort=None, title=None):
        self._sort = sort
        self._sort_orig = sort

        # Create peer filter generator
        if pfilter is not None:
            def filter_peers(peers):
                yield from pfilter.apply(peers)
        else:
            def filter_peers(peers):
                yield from peers
        self._maybe_filter_peers = filter_peers

        self._title = title
        self.title_updater = None

        self._peers = ()
        self._initialized = False

        self._table = Table(**TUICOLUMNS)
        self._table.columns = columns

        self._listbox = tui.keymap.wrap(urwid.ListBox, context='peerlist')(urwid.SimpleListWalker([]))
        listbox_sb = urwid.AttrMap(
            ScrollBar(urwid.AttrMap(self._listbox, 'peerlist')),
            'scrollbar'
        )
        pile = urwid.Pile([
            ('pack', urwid.AttrMap(self._table.headers, 'peerlist.header')),
            listbox_sb
        ])
        super().__init__(pile)

        self._poller = srvapi.create_poller(
            srvapi.torrent.torrents, tfilter, keys=('peers', 'name', 'id'))
        self._poller.on_response(self._handle_response)

    def _handle_response(self, response):
        if response is None or not response.torrents:
            self.clear()
        else:
            def peers_combined(torrents):
                for t in torrents:
                    yield from self._maybe_filter_peers(t['peers'])
            self._peers = {p['id']:p for p in peers_combined(response.torrents)}

        if self.title_updater is not None:
            # First argument can be cropped if too long, second argument is fixed
            self.title_updater(self.title, ' [%d]' % self.count)

        self._invalidate()

    def render(self, size, focus=False):
        if self._peers is not None:
            self._update_listitems()
            self._peers = None
        return super().render(size, focus)

    def _update_listitems(self):
        pdict = self._peers
        walker = self._listbox.body
        dead_pws = []
        for pw in walker:  # pw = PeerListItemWidget
            pid = pw.pid
            try:
                # Update existing peer widget with new data
                pw.update(pdict[pid])
                del pdict[pid]
            except KeyError:
                # Peer no longer exists
                dead_pws.append(pw)

        # Remove list items
        for pw in dead_pws:
            walker.remove(pw)

        # Any peers that haven't been used to update an existing peer widget are new
        for pid in pdict:
            self._table.register(pid)
            row = urwid.AttrMap(self._table.get_row(pid), attr_map='peerlist')
            walker.append(PeerListItemWidget(pdict[pid], row))

        # Sort peers
        if self._sort is not None:
            self._sort.apply(walker,
                            item_getter=lambda pw: pw.peer,
                            inplace=True)

    def clear(self):
        """Remove all list items"""
        self._table.clear()
        self._listbox.body[:] = []
        self._listbox._invalidate()

    @property
    def sort(self):
        """TorrentPeerSorter object (set to `None` to restore original sort order)"""
        return self._sort

    @sort.setter
    def sort(self, sort):
        if sort == 'RESET':
            self._sort = self._sort_orig
        else:
            self._sort = sort
        self._poller.poll()

    @property
    def title(self):
        # Create the fixed part of the title (everything minus the number of peers listed)
        # If title is not given, create one from filter and sort order
        if self._title is None:
            # tfilter is either None or an actual TorrentFilter instance
            title = str(tfilter or 'all')
            if pfilter:
                title += ' %s' % pfilter
        else:
            title = self._title

        if self._sort is not None:
            sortstr = str(self._sort)
            if sortstr is not self._sort.DEFAULT_SORT:
                title += ' {%s}' % sortstr
        return title

    @property
    def count(self):
        """Number of listed peers"""
        # If this method was called before rendering, the contents of the
        # listbox widget are inaccurate and we have to use self._peers.  But
        # if we're called after rendering, self._peers is reset to None.
        if self._peers is not None:
            return len(self._peers)
        else:
            return len(self._listbox.body)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""TUI and CLI specs for torrent list columns"""

from ..logging import make_logger
log = make_logger(__name__)

from . import ColumnBase

COLUMNS = {}

import os
PATHSEP = os.sep
class Path(ColumnBase):
    header = {'left': 'Path'}
    width = None
    align = 'left'
    needed_keys = ('path',)

    def get_value(self):
        path = self.data['path'].rstrip(PATHSEP)
        if self.width is None:
            return path

        # Remove characters from the end of the most toplevel directory
        # until there's only one left. Then do the same with the next
        # level.
        # /the/path/to/your/torrents
        # -> /th/path/to/your/torrents
        # -> /t/path/to/your/torrents
        # -> /t/pat/to/your/torrents
        while len(path) > self.width:
            dirs = [d for d in path.split(PATHSEP) if d]
            for i,d in enumerate(dirs):
                if len(d) > 1:
                    dirs[i] = d[:-1]
                    break
            path = PATHSEP + PATHSEP.join(dirs)
            if len(path) <= len(dirs)*2:
                break  # Each dir is now exactly one character long.

        # If "/t/p/t/y/t" is still too long, simply remove enough characters
        # from the front.
        if len(path) > self.width:
            excess = len(path) - self.width + 1
            path = '…' + path[excess:]

        return path

COLUMNS['path'] = Path


class Connections(ColumnBase):
    header = {'left': 'Conn'}
    width = 5
    needed_keys = ('peers-connected',)

    def get_value(self):
        return self.data['peers-connected']

    def get_raw(self):
        return int(self.get_value())

COLUMNS['connections'] = Connections


class Seeds(ColumnBase):
    header = {'left': 'Seeds'}
    width = 5
    needed_keys = ('peers-seeding',)

    def get_value(self):
        return self.data['peers-seeding']

    def get_raw(self):
        return int(self.get_value())

COLUMNS['seeds'] = Seeds


class Progress(ColumnBase):
    header = {'right': '%'}
    width = 4
    needed_keys = ('%verified', '%downloaded', '%metadata')

    def get_value(self):
        t = self.data
        v, d, m = (t['%verified'], t['%downloaded'], t['%metadata'])
        if 0 < v < 100:
            return v
        elif 0 < m < 100:
            return m
        else:
            return d

    def get_raw(self):
        return float(self.get_value()) / 100

COLUMNS['progress'] = Progress


class PercentAvailable(ColumnBase):
    header = {'left': 'Avail', 'right': '%'}
    width = 7
    needed_keys = ('%available', 'peers-seeding')

    def get_value(self):
        data = self.data
        if data['peers-seeding'] > 0:
            return type(data['%available'])(100)
        else:
            return data['%available']

    def get_raw(self):
        return float(self.get_value()) / 100

COLUMNS['%available'] = PercentAvailable


class Ratio(ColumnBase):
    header = {'left': 'Ratio'}
    width = 5
    needed_keys = ('ratio',)

    def get_value(self):
        return self.data['ratio']

    def get_raw(self):
        return float(self.get_value())

COLUMNS['ratio'] = Ratio


class Size(ColumnBase):
    header = {'left': 'Size', 'right': '?'}
    width = 6
    needed_keys = ('size-final',)

    def get_value(self):
        return self.data['size-final']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['size'] = Size


class Downloaded(ColumnBase):
    header = {'left': 'Dn', 'right': '?'}
    width = 6
    needed_keys = ('size-downloaded', 'size-final')

    def get_value(self):
        return self.data['size-downloaded']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['downloaded'] = Downloaded


class Uploaded(ColumnBase):
    header = {'left': 'Up', 'right': '?'}
    width = 6
    needed_keys = ('size-uploaded', 'size-downloaded')

    def get_value(self):
        return self.data['size-uploaded']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['uploaded'] = Uploaded


class BytesAvailable(ColumnBase):
    header = {'left': 'Avail', 'right': '?'}
    width = 7
    needed_keys = ('size-available', 'size-final', 'peers-seeding')

    def get_value(self):
        data = self.data
        if data['peers-seeding'] > 0:
            return data['size-final']
        else:
            return data['size-available']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = unit

COLUMNS['available'] = BytesAvailable


class RateDown(ColumnBase):
    header = {'left': 'Dn', 'right': '?/s'}
    width = 6
    needed_keys = ('rate-down',)

    def get_value(self):
        return self.data['rate-down']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-down'] = RateDown


class RateUp(ColumnBase):
    header = {'left': 'Up', 'right': '?/s'}
    width = 6
    needed_keys = ('rate-up',)

    def get_value(self):
        return self.data['rate-up']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-up'] = RateUp


class RateLimitDown(ColumnBase):
    header = {'left': 'LmtDn', 'right': '?/s'}
    width = 9
    needed_keys = ('rate-limit-down',)

    def get_value(self):
        return self.data['rate-limit-down']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-limit-down'] = RateLimitDown


class RateLimitUp(ColumnBase):
    header = {'left': 'LmtUp', 'right': '?/s'}
    width = 9
    needed_keys = ('rate-limit-up',)

    def get_value(self):
        return self.data['rate-limit-up']

    def get_raw(self):
        return int(self.get_value())

    @classmethod
    def set_unit(cls, unit):
        cls.header['right'] = '%s/s' % unit

COLUMNS['rate-limit-up'] = RateLimitUp


class EtaComplete(ColumnBase):
    header = {'left': 'ETA'}
    width = 5
    needed_keys = ('timespan-eta',)

    def get_value(self):
        return self.data['timespan-eta']

    def get_raw(self):
        return int(self.get_value())

COLUMNS['eta'] = EtaComplete


class TorrentName(ColumnBase):
    header = {'left': 'Name'}
    width = None
    needed_keys = ('name',)
    align = 'left'
    may_have_wide_chars = True

    def get_value(self):
        return self.data['name']

COLUMNS['name'] = TorrentName


class Status(ColumnBase):
    header = {'left': 'Status'}
    width = 11
    needed_keys = ('status',)

    def get_value(self):
        return self.data['status'][0]

COLUMNS['status'] = Status


class Tracker(ColumnBase):
    header = {'left': 'Tracker'}
    width = 10
    needed_keys = ('trackers',)
    align = 'left'

    def get_value(self):
        if len(self.data['trackers']) > 0:
            return self.data['trackers'][0]['url-announce'].domain
        else:
            return ''

COLUMNS['tracker'] = Tracker


class Error(ColumnBase):
    header = {'left': 'Error'}
    width = ('weight', 300)
    needed_keys = ('error',)
    align = 'left'

    def get_value(self):
        return self.data['error']

COLUMNS['error'] = Error


class Marked(ColumnBase):
    interfaces = ('tui',)

COLUMNS['marked'] = Marked


class TimeBase(ColumnBase):
    width = 10

    def get_raw(self):
        return int(self.get_value())

class TimeCreated(TimeBase):
    header = {'left': 'Created'}
    needed_keys = ('time-created',)

    def get_value(self):
        return self.data['time-created']

COLUMNS['created'] = TimeCreated

class TimeAdded(TimeBase):
    header = {'left': 'Added'}
    needed_keys = ('time-added',)

    def get_value(self):
        return self.data['time-added']

COLUMNS['added'] = TimeAdded

class TimeStarted(TimeBase):
    header = {'left': 'Started'}
    needed_keys = ('time-started',)

    def get_value(self):
        return self.data['time-started']

COLUMNS['started'] = TimeStarted

class TimeActivity(TimeBase):
    header = {'left': 'Activity'}
    needed_keys = ('time-activity',)

    def get_value(self):
        return self.data['time-activity']

COLUMNS['activity'] = TimeActivity

class TimeCompleted(TimeBase):
    header = {'left': 'Completed'}
    needed_keys = ('time-completed',)

    def get_value(self):
        return self.data['time-completed']

COLUMNS['completed'] = TimeCompleted



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import blinker
import operator
from functools import reduce
from collections import abc

from .poll import RequestPoller
from .filters.tfilter import TorrentFilter


class TorrentRequestPool():
    """Combine multiple `TorrentAPI.torrents` requests into one

    The wanted Torrent keys from all subscribers are combined and added to the
    needed keys for TorrentFilter from all subscribers.

    After the combined torrents have arrived, split it back up by using each
    subscriber's filter and provide it to its callbacks as tuples.
    """
    def __init__(self, srvapi, interval=1):
        self._api = srvapi.torrent
        self._tfilters = {}
        self._keys = {}
        self._poller = RequestPoller(request=self._api.torrents,
                                     autoconnect=False,
                                     interval=interval,
                                     loop=srvapi.loop)
        self._poller.on_response(self._handle_tlist)

    def register(self, sid, callback, keys=(), tfilter=None):
        """Add new request to request pool

        sid: Subscriber ID (any hashable)
        callback: Callable that receives a tuple of Torrents on updates
        keys: Wanted Torrent keys
        tfilter: None for all torrents or TorrentFilter instance
        """
        if isinstance(tfilter, abc.Sequence):
            tfilter = TorrentFilter('|'.join('id=%s' % tid for tid in tfilter))

        log.debug('Registering subscriber: %s', sid)
        event = blinker.signal(sid)
        event.connect(callback)
        self._keys[event] = tuple(keys)
        self._tfilters[event] = tfilter

        # It's possible that a currently ongoing request doesn't collect the
        # keys this new callback needs.  In that case, the request is finished
        # AFTER we added the callback, and the callback would be called with
        # lacking keys, resuling in a KeyError.
        # Therefore we ask the poller to dump the result of a currently
        # ongoing request to prevent this.
        if self._poller.running:
            self._poller.skip_ongoing_request()

        self._combine_requests()

    def _combine_requests(self):
        """Create single request that combines keys and filters of all subscribers"""
        if not self.has_subscribers:
            # Don't request anything
            log.debug('No subscribers - setting request to None')
            self._poller.set_request(None)
        else:
            kwargs = {}

            all_filters = tuple(self._tfilters.values())
            if not all_filters or None in all_filters:
                # No subscribers or at least one subscriber wants all torrents
                kwargs['torrents'] = None
            else:
                kwargs['torrents'] = reduce(operator.__add__, all_filters)

            kwargs['keys'] = reduce(operator.__add__, self._keys.values())
            # Filters also need certain keys
            for f in all_filters:
                if f is not None:
                    kwargs['keys'] += f.needed_keys

            kwargs['keys'] = tuple(set(kwargs['keys']))
            log.debug('Combined filters: %s', kwargs['torrents'])
            log.debug('Combined keys: %s', kwargs['keys'])
            self._poller.set_request(self._api.torrents, **kwargs)

    def _handle_tlist(self, response):
        # If the request failed, response is None and tlist is empty.
        tlist = response.torrents if response is not None else ()

        dead_subscribers = []
        def has_subscribers(event):
            if not bool(event.receivers):
                dead_subscribers.append(event.name)
                return False
            else:
                return True

        log.debug('Processing %d torrents for %d subscribers',
                  len(tlist), len(self._tfilters))
        if len(self._tfilters) == 1:
            # If there's only one subscriber, there's no need to filter the
            # torrents again.
            event = next(iter(self._tfilters))
            if has_subscribers(event):
                log.debug('Running callback: %r', event.name)
                event.send(tlist)
        else:
            # More than 1 subscriber means we have to filter the torrents
            # again for each one.
            for event,filter in self._tfilters.items():
                if has_subscribers(event):
                    log.debug('Running callback: %r', event.name)
                    if filter is None:
                        # Subscriber wants all torrents
                        this_tlist = tlist
                    else:
                        # Subscriber wants filtered torrents
                        this_tlist = filter.apply(tlist)
                    event.send(this_tlist)

        # Remove dead subscribers
        for eventname in dead_subscribers:
            self.remove(eventname)

    def remove(self, sid):
        """Unsubscribe previously registered subscriber"""
        log.debug('Removing subscriber: %s', sid)
        event = blinker.signal(sid)
        del self._keys[event]
        del self._tfilters[event]
        self._combine_requests()

    @property
    def has_subscribers(self):
        """Whether any subscribers are registered"""
        return bool(self._tfilters)

    def __getattr__(self, attr):
        if attr in ('start', 'stop', 'poll', 'running'):
            return getattr(self._poller, attr)
        raise AttributeError('{!r} object has no attribute {!r}'
                             .format(type(self).__name__, attr))

    @property
    def interval(self):
        return self._poller.interval

    @interval.setter
    def interval(self, interval):
        self._poller.interval = interval



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)

import os
from appdirs import (user_cache_dir, user_config_dir)

from .. import APPNAME
from ..views.tlist import COLUMNS as TCOLUMNS
from ..views.flist import COLUMNS as FCOLUMNS
from ..views.plist import COLUMNS as PCOLUMNS
from ..client.sorters.tsorter import TorrentSorter
from ..client.sorters.psorter import TorrentPeerSorter

DEFAULT_RCFILE        = user_config_dir(APPNAME) + '/rc'
DEFAULT_HISTORY_FILE  = user_cache_dir(APPNAME)+'/history'
DEFAULT_THEME_FILE    = os.path.join(os.path.dirname(__file__), 'default.theme')

DEFAULT_TLIST_SORT    = ('name',)
DEFAULT_PLIST_SORT    = ('torrent',)
DEFAULT_TLIST_COLUMNS = ('marked', 'size', 'downloaded', 'uploaded', 'ratio',
                         'seeds', 'connections', 'status', 'eta', 'progress',
                         'rate-down', 'rate-up', 'name')
DEFAULT_FLIST_COLUMNS = ('marked', 'priority', 'progress', 'downloaded', 'size', 'name')

from ..client.geoip import GEOIP_AVAILABLE
if GEOIP_AVAILABLE:
    DEFAULT_PLIST_COLUMNS = ('progress', 'rate-down', 'rate-up', 'rate-est',
                             'eta', 'ip', 'country', 'client')
else:
    DEFAULT_PLIST_COLUMNS = ('progress', 'rate-down', 'rate-up', 'rate-est',
                             'eta', 'ip', 'client')


from .settings import (StringValue, IntegerValue, NumberValue, BooleanValue,
                       PathValue, ListValue, SetValue, OptionValue)

class SortOrderValue(SetValue):
    """SetValue that correctly validates inverted sort orders (e.g. '!name')"""
    def __init__(self, sortercls, *args, options=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.options = tuple(sortercls.SORTSPECS)

    def validate(self, names):
        super().validate(name.strip('!.') for name in names)


def init_defaults(cfg):
    cfg.load(
        StringValue('srv.url', default='localhost:9091',
                    description='URL of the Transmission RPC interface ([USER:PASSWORD@]HOST[:PORT])'),
        NumberValue('srv.timeout', default=10, min=0,
                    description=('Number of seconds before connecting '
                                 'to Transmission daemon fails')),

        SetValue('columns.torrents', default=DEFAULT_TLIST_COLUMNS,
                 options=tuple(TCOLUMNS),
                 description='List of columns in new torrent lists'),
        SetValue('columns.peers', default=DEFAULT_PLIST_COLUMNS,
                 options=tuple(PCOLUMNS),
                 description='List of columns in new peer lists'),
        SetValue('columns.files', default=DEFAULT_FLIST_COLUMNS,
                 options=tuple(FCOLUMNS),
                 description='List of columns in new torrent file lists'),

        SortOrderValue(TorrentSorter, 'sort.torrents', default=DEFAULT_TLIST_SORT,
                       description='List of torrent list sort orders'),
        SortOrderValue(TorrentPeerSorter, 'sort.peers', default=DEFAULT_PLIST_SORT,
                       description='List of peer list sort orders'),

        PathValue('tui.theme', default=DEFAULT_THEME_FILE,
                  description='Path to theme file'),
        IntegerValue('tui.log.height', default=10, min=1,
                     description='Maximum height of the log section'),
        NumberValue('tui.log.autohide', default=10, min=0,
                    description=('If the log is hidden, show it for this many seconds '
                                 'for new log entries before hiding it again')),
        PathValue('tui.cli.history', default=DEFAULT_HISTORY_FILE,
                  description='Path to TUI command line history file'),
        NumberValue('tui.poll', default=5, min=0.1,
                    description='Interval in seconds between TUI updates'),

        OptionValue('unit.bandwidth', default='byte', options=('bit', 'byte'),
                    description="Unit for bandwidth rates ('bit' or 'byte')"),
        OptionValue('unitprefix.bandwidth', default='metric', options=('metric', 'binary'),
                    description=("Unit prefix for bandwidth rates ('metric' or 'binary')")),

        OptionValue('unit.size', default='byte', options=('bit', 'byte'),
                    description="Unit for sizes ('bit' or 'byte')"),
        OptionValue('unitprefix.size', default='binary', options=('metric', 'binary'),
                    description=("Unit prefix for sizes ('metric' or 'binary')")),

        StringValue('tui.marked.on', default='✔', minlen=1, maxlen=1,
                    description=('Character displayed in "marked" column for marked '
                                 'list items (see "mark" command)')),
        StringValue('tui.marked.off', default=' ', minlen=1, maxlen=1,
                    description=('Character displayed in "marked" column for unmarked '
                                 'list items (see "mark" command)')),
    )


from .settings_server import (RateLimitSrvValue, PathSrvValue, PathIncompleteSrvValue)

def init_server_defaults(cfg, settingsapi):
    cfg.load(
        RateLimitSrvValue('srv.limit.rate.up',
                          description='Combined upload rate limit',
                          getter=lambda: settingsapi['rate-limit-up'],
                          setter=settingsapi.set_rate_limit_up),
        RateLimitSrvValue('srv.limit.rate.down',
                          description='Combined download rate limit',
                          getter=lambda: settingsapi['rate-limit-down'],
                          setter=settingsapi.set_rate_limit_down),

        PathSrvValue('srv.path.complete',
                     description='Where to put torrent files',
                     getter=lambda: settingsapi['path-complete'],
                     setter=settingsapi.set_path_complete),

        PathIncompleteSrvValue('srv.path.incomplete',
                     description='Where to put incomplete torrent files',
                     getter=lambda: settingsapi['path-incomplete'],
                     setter=settingsapi.set_path_incomplete),
    )


DEFAULT_KEYMAP = (
    # Use some vi and emacs keybindings
    {'context': None, 'key': 'h', 'action': '<left>'},
    {'context': None, 'key': 'j', 'action': '<down>'},
    {'context': None, 'key': 'k', 'action': '<up>'},
    {'context': None, 'key': 'l', 'action': '<right>'},
    {'context': None, 'key': 'g', 'action': '<home>'},
    {'context': None, 'key': 'G', 'action': '<end>'},
    {'context': None, 'key': 'ctrl-n', 'action': '<down>'},
    {'context': None, 'key': 'ctrl-p', 'action': '<up>'},
    {'context': None, 'key': 'ctrl-f', 'action': '<right>'},
    {'context': None, 'key': 'ctrl-b', 'action': '<left>'},

    # Global TUI keys
    {'context': 'main', 'key': 'q', 'action': 'quit'},
    {'context': 'main', 'key': ':', 'action': 'tui show cli'},

    {'context': 'main', 'key': 'F1+c', 'action': 'tab help commands'},
    {'context': 'main', 'key': 'F1+s', 'action': 'tab help settings'},
    {'context': 'main', 'key': 'F1+k', 'action': 'tab help keymap'},
    {'context': 'main', 'key': 'F1+f', 'action': 'tab help filtering'},
    {'context': 'main', 'key': 'F1+r', 'action': 'tab help rcfile'},
    {'context': 'main', 'key': '?',    'action': '<F1>'},

    {'context': 'main', 'key': 'ctrl-l', 'action': 'clearlog'},
    {'context': 'main', 'key': 'meta-L', 'action': 'tui toggle log'},
    {'context': 'main', 'key': 'meta-M', 'action': 'tui toggle main'},
    {'context': 'main', 'key': 'meta-T', 'action': 'tui toggle topbar'},
    {'context': 'main', 'key': 'meta-B', 'action': 'tui toggle bottombar'},

    # Bandwidth limits
    {'context': 'main', 'key': 'shift-up',    'action': 'set srv.limit.rate.down +=100kB'},
    {'context': 'main', 'key': 'shift-down',  'action': 'set srv.limit.rate.down -=100kB'},
    {'context': 'main', 'key': 'shift-right', 'action': 'set srv.limit.rate.up +=100kB'},
    {'context': 'main', 'key': 'shift-left',  'action': 'set srv.limit.rate.up -=100kB'},

    # Tab management
    {'context': 'tabs', 'key': 'n', 'action': 'tab'},
    {'context': 'tabs', 'key': 'd', 'action': 'tab --close'},
    {'context': 'tabs', 'key': 'D', 'action': 'tab --close --focus left'},
    {'context': 'tabs', 'key': 'meta-1', 'action': 'tab --focus 1'},
    {'context': 'tabs', 'key': 'meta-2', 'action': 'tab --focus 2'},
    {'context': 'tabs', 'key': 'meta-3', 'action': 'tab --focus 3'},
    {'context': 'tabs', 'key': 'meta-4', 'action': 'tab --focus 4'},
    {'context': 'tabs', 'key': 'meta-5', 'action': 'tab --focus 5'},
    {'context': 'tabs', 'key': 'meta-6', 'action': 'tab --focus 6'},
    {'context': 'tabs', 'key': 'meta-7', 'action': 'tab --focus 7'},
    {'context': 'tabs', 'key': 'meta-8', 'action': 'tab --focus 8'},
    {'context': 'tabs', 'key': 'meta-9', 'action': 'tab --focus 9'},
    {'context': 'tabs', 'key': 'meta-0', 'action': 'tab --focus 10'},

    # List torrents with different filters
    {'context': 'tabs', 'key': 'f a', 'action': 'tab ls active'},
    {'context': 'tabs', 'key': 'f A', 'action': 'tab ls !active'},
    {'context': 'tabs', 'key': 'f i', 'action': 'tab ls isolated --columns name,tracker,error'},
    {'context': 'tabs', 'key': 'f p', 'action': 'tab ls paused'},
    {'context': 'tabs', 'key': 'f P', 'action': 'tab ls !paused'},
    {'context': 'tabs', 'key': 'f c', 'action': 'tab ls complete'},
    {'context': 'tabs', 'key': 'f C', 'action': 'tab ls !complete'},
    {'context': 'tabs', 'key': 'f u', 'action': 'tab ls uploading'},
    {'context': 'tabs', 'key': 'f d', 'action': 'tab ls downloading'},
    {'context': 'tabs', 'key': 'f s', 'action': 'tab ls seeding'},
    {'context': 'tabs', 'key': 'f l', 'action': 'tab ls leeching'},
    {'context': 'tabs', 'key': 'f .', 'action': 'tab ls'},

    # Torrent list actions
    {'context': 'torrentlist', 'key': 's d',       'action': 'sort --add dir'},
    {'context': 'torrentlist', 'key': 's D',       'action': 'sort --add !dir'},
    {'context': 'torrentlist', 'key': 's e',       'action': 'sort --add eta'},
    {'context': 'torrentlist', 'key': 's E',       'action': 'sort --add !eta'},
    {'context': 'torrentlist', 'key': 's n',       'action': 'sort --add name'},
    {'context': 'torrentlist', 'key': 's N',       'action': 'sort --add !name'},
    {'context': 'torrentlist', 'key': 's o',       'action': 'sort --add ratio'},
    {'context': 'torrentlist', 'key': 's O',       'action': 'sort --add !ratio'},
    {'context': 'torrentlist', 'key': 's p',       'action': 'sort --add progress'},
    {'context': 'torrentlist', 'key': 's P',       'action': 'sort --add !progress'},
    {'context': 'torrentlist', 'key': 's r',       'action': 'sort --add rate'},
    {'context': 'torrentlist', 'key': 's R',       'action': 'sort --add !rate'},
    {'context': 'torrentlist', 'key': 's s',       'action': 'sort --add seeds'},
    {'context': 'torrentlist', 'key': 's S',       'action': 'sort --add !seeds'},
    {'context': 'torrentlist', 'key': 's t',       'action': 'sort --add tracker'},
    {'context': 'torrentlist', 'key': 's T',       'action': 'sort --add !tracker'},
    {'context': 'torrentlist', 'key': 's z',       'action': 'sort --add size'},
    {'context': 'torrentlist', 'key': 's Z',       'action': 'sort --add !size'},
    {'context': 'torrentlist', 'key': 's ,',       'action': 'sort --reset'},
    {'context': 'torrentlist', 'key': 's .',       'action': 'sort --none'},

    # Torrent actions
    {'context': 'torrent', 'key': 't v',       'action': 'verify'},
    {'context': 'torrent', 'key': 't p',       'action': 'pause --toggle'},
    {'context': 'torrent', 'key': 't P',       'action': 'start --toggle --force'},
    {'context': 'torrent', 'key': 't a',       'action': 'announce'},
    {'context': 'torrent', 'key': 't d',       'action': 'delete'},
    {'context': 'torrent', 'key': 't D',       'action': 'delete --delete-files'},
    {'context': 'torrent', 'key': 'enter',     'action': 'tab details'},
    {'context': 'torrent', 'key': 'alt-enter', 'action': 'tab filelist'},
    {'context': 'torrent', 'key': 'space',     'action': 'mark --toggle --focus-next'},
    {'context': 'torrent', 'key': 'alt-space', 'action': 'mark --toggle --all'},

    # Peer list actions
    {'context': 'peerlist', 'key': 's c',       'action': 'sort --add country'},
    {'context': 'peerlist', 'key': 's C',       'action': 'sort --add !country'},
    {'context': 'peerlist', 'key': 's d',       'action': 'sort --add rate-down'},
    {'context': 'peerlist', 'key': 's D',       'action': 'sort --add !rate-down'},
    {'context': 'peerlist', 'key': 's e',       'action': 'sort --add eta'},
    {'context': 'peerlist', 'key': 's E',       'action': 'sort --add !eta'},
    {'context': 'peerlist', 'key': 's l',       'action': 'sort --add client'},
    {'context': 'peerlist', 'key': 's L',       'action': 'sort --add !client'},
    {'context': 'peerlist', 'key': 's p',       'action': 'sort --add progress'},
    {'context': 'peerlist', 'key': 's P',       'action': 'sort --add !progress'},
    {'context': 'peerlist', 'key': 's u',       'action': 'sort --add rate-up'},
    {'context': 'peerlist', 'key': 's U',       'action': 'sort --add !rate-up'},
    {'context': 'peerlist', 'key': 's s',       'action': 'sort --add rate-est'},
    {'context': 'peerlist', 'key': 's S',       'action': 'sort --add !rate-est'},
    {'context': 'peerlist', 'key': 's r',       'action': 'sort --add rate'},
    {'context': 'peerlist', 'key': 's R',       'action': 'sort --add !rate'},
    {'context': 'peerlist', 'key': 's t',       'action': 'sort --add torrent'},
    {'context': 'peerlist', 'key': 's T',       'action': 'sort --add !torrent'},
    {'context': 'peerlist', 'key': 's ,',       'action': 'sort --reset'},
    {'context': 'peerlist', 'key': 's .',       'action': 'sort --none'},

    # Torrent file actions
    {'context': 'file', 'key': '+',         'action': 'priority high'},
    {'context': 'file', 'key': '=',         'action': 'priority normal'},
    {'context': 'file', 'key': '-',         'action': 'priority low'},
    {'context': 'file', 'key': '0',         'action': 'priority shun'},
    {'context': 'file', 'key': 'space',     'action': 'mark --toggle --focus-next'},
    {'context': 'file', 'key': 'alt-space', 'action': 'mark --toggle --all'},
)



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

from ..logging import make_logger
log = make_logger(__name__)


class CallbackDict(dict):
    """Mapping that runs a callback on every change"""
    # https://stackoverflow.com/a/5186698

    __slots__ = ["callback"]

    def __init__(self, callback, *args, **kwargs):
        self.callback = callback
        dict.__init__(self, *args, **kwargs)

    def _wrap(method):
        def wrapper(self, *args, **kwargs):
            result = method(self, *args, **kwargs)
            self.callback()
            return result
        return wrapper

    __delitem__ = _wrap(dict.__delitem__)
    __setitem__ = _wrap(dict.__setitem__)
    clear = _wrap(dict.clear)
    pop = _wrap(dict.pop)
    popitem = _wrap(dict.popitem)
    setdefault = _wrap(dict.setdefault)
    update =  _wrap(dict.update)


from collections import abc
def listify_args(args):
    """Make list from `args`

    If `args` is a string, use split(',') on it.  If `args` is any other
    iterable, recurse on each item and add the return values to the result.
    Otherwise make `args` a single item.

    All items are turned into strings and strip()ed.

    Despite the name, return a tuple.
    """
    if args is None:
        return []

    if isinstance(args, str):
        # args is a string of comma-separated list items
        args = (str(arg).strip() for arg in args.split(','))
    elif isinstance(args, abc.Iterable):
        # args is an iterable of list items; each item may be a string of
        # comma-separated items
        args = (a for arg in args
                for a in listify_args(arg))
    else:
        # args is something else
        args = (str(args).strip())
    return [arg for arg in args if arg]


def log_msgs(logger, msgs, quiet=False):
    """Log messages to `logger`

    `msgs` is an iterable of strings or exceptions.  Strings are logged to
    level INFO, exceptions are logged to level ERROR.

    If `quiet` evaluates to True, INFO messages are not logged.
    """
    for msg in msgs:
        if isinstance(msg, str):
            if not quiet:
                logger.info(msg)
        elif isinstance(msg, Exception):
            logger.error(msg)
        else:
            raise RuntimeError('Invalid log message type ({}): {!r}'
                               .format(type(msg).__name__, msg))



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

import urwid

from ..table import ColumnHeaderWidget
from . import (Style, CellWidgetBase)
from ...views.tlist import COLUMNS as _COLUMNS

TUICOLUMNS = {}

class MarkedBase(_COLUMNS['marked'], CellWidgetBase):
    width = 1
    needed_keys = ()
    _marked_char = '#'
    _unmarked_char = ' '

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.text.set_text(self._unmarked_char)
        self._is_marked = False

    @classmethod
    def set_marked_char(cls, char):
        cls._marked_char = str(char)
        cls.header.original_widget.left = cls._marked_char

    @classmethod
    def set_unmarked_char(cls, char):
        cls._unmarked_char = str(char)

    @property
    def is_marked(self):
        return self._is_marked

    @is_marked.setter
    def is_marked(self, is_marked):
        if is_marked:
            self.text.set_text(self._marked_char)
        else:
            self.text.set_text(self._unmarked_char)
        self._is_marked = is_marked

    def update(self, data):
        pass  # Ignore update data

    def __repr__(self):
        return '<{} {}>'.format(type(self).__name__, 'on' if self._is_marked else 'off')



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Monkey patches that should be removed when they are resolved upstream"""

import urwid

# Add more actions for key bindings
urwid.CURSOR_WORD_LEFT = 'cursor word left'
urwid.CURSOR_WORD_RIGHT = 'cursor word right'
urwid.DELETE_TO_EOL = 'delete to end of line'
urwid.DELETE_LINE = 'delete line'
urwid.DELETE_CHAR_UNDER_CURSOR = 'delete char under cursor'
urwid.DELETE_WORD_LEFT = 'delete word left'
urwid.DELETE_WORD_RIGHT = 'delete word right'
urwid.CANCEL = 'cancel'


# Limit the impact of the high CPU load bug
# https://github.com/urwid/urwid/pull/86
# https://github.com/urwid/urwid/issues/90
from distutils.version import LooseVersion
if LooseVersion(urwid.__version__) <= LooseVersion('1.3.1'):
    urwid.AsyncioEventLoop._idle_emulation_delay = 1/25


# Raise UnicodeDecodeError properly
# https://github.com/urwid/urwid/pull/92
# https://github.com/urwid/urwid/pull/196
class AsyncioEventLoop_patched(urwid.AsyncioEventLoop):
    def _exception_handler(self, loop, context):
        exc = context.get('exception')
        if exc:
            loop.stop()
            if not isinstance(exc, urwid.ExitMainLoop):
                self._exc_info = exc
        else:
            loop.default_exception_handler(context)

    def run(self):
        """
        Start the event loop.  Exit the loop when any callback raises
        an exception.  If ExitMainLoop is raised, exit cleanly.
        """
        self._loop.set_exception_handler(self._exception_handler)
        self._loop.run_forever()
        if self._exc_info:
            raise self._exc_info

urwid.AsyncioEventLoop = AsyncioEventLoop_patched


class ListBox_patched(urwid.ListBox):
    def keypress(self, size, key):
        # Offer key to focused widget first
        # https://github.com/urwid/urwid/pull/233
        focused_widget = self.focus
        if focused_widget is not None and focused_widget.selectable() and \
           focused_widget.keypress((size[0],), key) is None:
            return None

        # Add support for home/end keys
        # https://github.com/urwid/urwid/pull/229
        key = super().keypress(size, key)
        if self.focus is not None:
            if key == 'home':
                self.focus_position = next(iter(self.body.positions()))
                self.set_focus_valign('top')
                self._invalidate()
                return None
            elif key == 'end':
                self.focus_position = next(iter(self.body.positions(reverse=True)))
                self.set_focus_valign('bottom')
                self._invalidate()
                return None
        return key


    # Add support for ScrollBar class (see stig.tui.scroll)
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._rows_max = None

    def _invalidate(self):
        super()._invalidate()
        self._rows_max = None

    def get_scrollpos(self, size, focus=False):
        """Current scrolling position

        Lower limit is 0, upper limit is the highest index of `body`.
        """
        middle, top, bottom = self.calculate_visible(size, focus)
        if middle is None:
            return 0
        else:
            offset_rows, _, focus_pos, _, _ = middle
            maxcol, maxrow = size
            flow_size = (maxcol,)

            body = self.body
            if hasattr(body, 'positions'):
                # For body[pos], pos can be anything, not just an int.  In that
                # case, the positions() method returns an interable of valid
                # positions.
                positions = tuple(self.body.positions())
                focus_index = positions.index(focus_pos)
                widgets_above_focus = (body[pos] for pos in positions[:focus_index])
            else:
                # Treat body like a normal list
                widgets_above_focus = (w for w in body[:focus_pos])

            rows_above_focus = sum(w.rows(flow_size) for w in widgets_above_focus)
            rows_above_top = rows_above_focus - offset_rows
            return rows_above_top

    def rows_max(self, size, focus=False):
        if self._rows_max is None:
            flow_size = (size[0],)
            body = self.body
            if hasattr(body, 'positions'):
                self._rows_max = sum(body[pos].rows(flow_size) for pos in body.positions())
            else:
                self._rows_max = sum(w.rows(flow_size) for w in self.body)
        return self._rows_max

urwid.ListBox = ListBox_patched



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Documentation commands"""

from ...logging import make_logger
log = make_logger(__name__)


from ..base import help as base
from .. import ExpectedResource
from .utils import make_tab_title_widget


class HelpCmd(base.HelpCmdbase):
    provides = {'tui'}
    tui = ExpectedResource

    def display_help(self, topics, lines):
        import urwid
        from ...tui.scroll import (Scrollable, ScrollBar)

        if hasattr(self, 'title'):
            titlew = make_tab_title_widget(str(self.title),
                                           attr_unfocused='tabs.help.unfocused',
                                           attr_focused='tabs.help.focused')
        else:
            titlew = make_tab_title_widget(','.join(topics),
                                           attr_unfocused='tabs.help.unfocused',
                                           attr_focused='tabs.help.focused')

        textw = urwid.AttrMap(Scrollable(urwid.Text('\n'.join(lines))), 'helptext')
        contentw = urwid.AttrMap(ScrollBar(textw), 'scrollbar')
        self.tui.tabs.load(titlew, contentw)


class VersionCmd(base.VersionCmdbase):
    provides = {'tui'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Documentation commands"""

from ...logging import make_logger
log = make_logger(__name__)

from ..base import help as base


class HelpCmd(base.HelpCmdbase):
    provides = {'cli'}

    async def run(self, TOPIC):
        from ...settings import is_server_setting
        for topic in TOPIC:
            if is_server_setting(topic):
                from ...main import srvapi
                from ...client import ClientError
                try:
                    await srvapi.settings.update()
                except ClientError as e:
                    log.error(str(e))
                finally:
                    break
        return super().run(TOPIC)

    def display_help(self, topics, lines):
        for line in lines:
            log.info(line)


class VersionCmd(base.VersionCmdbase):
    provides = {'cli'}



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Mixin classes for CLI commands"""

from ...logging import make_logger
log = make_logger(__name__)

from .. import ExpectedResource
from .. import utils


class make_request():
    async def make_request(self, request_coro, polling_frenzy=False, quiet=False):
        """Awaits request coroutine and logs messages; returns response"""
        response = await request_coro
        utils.log_msgs(log, response.msgs, quiet)
        return response


class select_torrents():
    def select_torrents(self, FILTER, allow_no_filter=True, discover_torrent=None):
        """Get TorrentFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentFilter and the
        resulting object is returned.

        If `FILTER` evaluates to False, None is returned if allow_no_filter
        evaluates to True, otherwise a ValueError is raised.

        `discover_torrent` is ignored and only used in the TUI version of this
        method (see ..tui.mixin.select_torrent).
        """
        if FILTER:
            from ...client import TorrentFilter
            return TorrentFilter(FILTER)
        else:
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent specified')


class select_files():
    def select_files(self, FILTER, allow_no_filter=True, discover_file=None):
        """Get TorrentFileFilter instance or None

        If `FILTER` evaluates to True, it is passed to TorrentFileFilter and
        the resulting object is returned.

        If `FILTER` evaluates to False, None is returned if allow_no_filter
        evaluates to True, otherwise a ValueError is raised.

        `discover_file` is ignored and only used in the TUI version of this
        method (see ..tui.mixin.select_file).
        """
        if FILTER:
            from ...client import TorrentFileFilter
            return TorrentFileFilter(FILTER)
        else:
            if allow_no_filter:
                return None
            else:
                raise ValueError('No torrent specified')


class only_supported_columns():
    def only_supported_columns(self, columns, specs):
        """Remove columns from list `columns` that don't support the CLI"""
        return [col for col in columns
                if 'cli' in specs[col].interfaces]



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt

"""Base classes for commands that provide CLI and TUI"""



# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details
# http://www.gnu.org/licenses/gpl-3.0.txt


# Only needed until asyncio gets a timeout again.
# https://github.com/python/asyncio/commit/e39449787dedd839e31946915fa933a08955b667
from aiohttp import Timeout as AsyncIOTimeout
import asyncio
class SleepUneasy():
    """Asynchronous sleep() that can be aborted"""

    def __init__(self, loop):
        self.loop = loop
        self._interrupt = asyncio.Event(loop=self.loop)

    async def sleep(self, seconds=None):
        """Sleep for `seconds` or until `interrupt` is called"""
        self._interrupt.clear()
        try:
            with AsyncIOTimeout(seconds, loop=self.loop):
                await self._interrupt.wait()
        except asyncio.TimeoutError:
            pass  # Interval passed without interrupt
        finally:
            self._interrupt.clear()

    def interrupt(self):
        """Stop sleeping"""
        self._interrupt.set()


class PerfectInterval():
    """Remove processing time from intervals"""

    def __init__(self, loop):
        self._loop = loop
        self._started = loop.time()

    def __call__(self, seconds):
        now = self._loop.time()
        stop_at = int(now) + seconds
        diff = stop_at - now
        return diff


from types import SimpleNamespace
class Response(SimpleNamespace):
    """Response to an API call

    All API implementations should use this class to provide return values to
    API calls.

    success: Whether the call was a success
    msgs: Sequence of messages; either strings or ClientError exceptions

    Any other keyword arguments are made available as attributes.
    """
    def __init__(self, success=False, msgs=(), **kwargs):
        super().__init__(success=bool(success), msgs=tuple(msgs), **kwargs)


from urllib.parse import (urlsplit, urlunsplit)
class URL():
    """Provide the same attributes as `urllib.parse.urlsplit` plus 'domain'"""
    def __init__(self, url):
        url = urlsplit(url)
        for attr in ('scheme', 'netloc', 'path', 'query', 'fragment'):
            setattr(self, attr, getattr(url, attr))

        # Find domain
        if url.hostname.count('.') <= 1:
            setattr(self, 'domain', url.hostname)
        else:
            parts = url.hostname.split('.')
            setattr(self, 'domain', '.'.join(parts[-2:]))

        self._url = url

    def __str__(self):
        return urlunsplit(self._url)

    def __repr__(self):
        return '<URL %s>' % self


def pretty_float(n):
    """Format number with a reasonable amount of decimal places"""
    n_abs = round(abs(n), 2)
    n_abs_int = int(n_abs)
    if n_abs == 0:
        return '0'
    elif n_abs == n_abs_int:
        return '%.0f' % n
    elif n_abs < 10:
        return '%.2f' % n
    elif n_abs < 100:
        return '%.1f' % n
    else:
        return '%.0f' % n
